#!/usr/bin/env python3
#
#  __init__.py
"""
Core (offline) functionality for the apeye library.
"""
#
#  Copyright © 2020-2022 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Redistribution and use in source and binary forms, with or without modification,
#  are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright notice,
#        this list of conditions and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.
#      * Neither the name of the copyright holder nor the names of its contributors
#        may be used to endorse or promote products derived from this software without
#        specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
#  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Based on the "pathlib" module from CPython.
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#
#  Based on Slumber <https://slumber.readthedocs.io>
#  Copyright (c) 2011 Donald Stufft
#  Licensed under the 2-clause BSD License
#

# stdlib
import ipaddress
import os
import pathlib
import re
from operator import attrgetter
from typing import (
		TYPE_CHECKING,
		Any,
		Dict,
		Iterable,
		List,
		Mapping,
		NamedTuple,
		Optional,
		Tuple,
		Type,
		TypeVar,
		Union,
		cast
		)
from urllib.parse import ParseResult, parse_qs, urlencode, urlparse, urlunparse

# 3rd party
from domdf_python_tools.doctools import prettify_docstrings
from domdf_python_tools.typing import PathLike

# this package
from apeye_core import _tld

if TYPE_CHECKING:
	# stdlib
	from typing import NoReturn

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2020-2022 Dominic Davis-Foster"
__license__: str = "BSD 3-Clause License"
__version__: str = "1.0.0b2"
__email__: str = "dominic@davis-foster.co.uk"

__all__ = ["URL", "URLPath", "Domain", "URLType", "URLPathType"]

# NOTE: Use the relevant version numbers for apeye itself.

URLType = TypeVar("URLType", bound="URL")

URLPathType = TypeVar("URLPathType", bound="URLPath")
"""
.. versionadded:: 1.1.0
"""


@prettify_docstrings
class URLPath(pathlib.PurePosixPath):
	"""
	Represents the path part of a URL.

	Subclass of :class:`pathlib.PurePosixPath` that provides a subset of its methods.

	.. versionchanged:: 1.1.0

		Implemented :meth:`~.apeye.url.URLPath.is_absolute`, :meth:`~.apeye.url.URLPath.joinpath`,
		:meth:`~.apeye.url.URLPath.relative_to`, :meth:`~.pathlib.PurePath.match`,
		``anchor``, ``drive``, and support for rich comparisons (``<``, ``<=``, ``>`` and ``>=``),
		which previously raised :exc:`NotImplementedError`.

	.. latex:clearpage::
	"""

	def __str__(self) -> str:
		"""
		Return the string representation of the path, suitable for passing to system calls.
		"""

		try:
			return self._str  # type: ignore
		except AttributeError:
			self._str = self._format_parsed_parts(self._drv, self._root, self._parts) or ''  # type: ignore
			return self._str

	def __repr__(self):
		return super().__repr__()

	@classmethod
	def _format_parsed_parts(cls, drv, root, parts):
		if drv or root:
			return drv + root + pathlib._posix_flavour.join(parts[1:])  # type: ignore
		else:
			return pathlib._posix_flavour.join(parts)  # type: ignore

	def is_absolute(self) -> bool:
		"""
		Returns whether the path is absolute (i.e. starts with ``/``).

		.. versionadded:: 1.1.0  previously raised :exc:`NotImplementedError`.
		"""

		return self.root == '/'

	def joinpath(self: URLPathType, *args) -> URLPathType:
		"""
		Combine this :class:`~.apeye.url.URLPath` with one or several arguments.

		.. versionadded:: 1.1.0  previously raised :exc:`NotImplementedError`.

		:returns: A new :class:`~.apeye.url.URLPath` representing either a subpath
			(if all arguments are relative paths) or a totally different path
			(if one of the arguments is absolute).
		"""

		return super().joinpath(*args)

	def relative_to(self: URLPathType, *other: PathLike) -> URLPathType:
		r"""
		Returns the relative path to another path identified by the passed arguments.

		The arguments are joined together to form a single path, and therefore the following behave identically:

		.. code-block:: pycon

			>>> URLPath("/news/sport").relative_to("/", "news")
			URLPath('sport')
			>>> URLPath("/news/sport").relative_to("/news")
			URLPath('sport')

		.. versionadded:: 1.1.0  previously raised :exc:`NotImplementedError`.

		:param \*other:

		:raises ValueError: if the operation is not possible (because this is not a subpath of the other path)

		.. latex:clearpage::

		.. seealso::

			:meth:`~.apeye.url.URL.relative_to`, which is recommended when constructing a relative path from a :class:`~URL`.
			This method cannot correctly handle some cases, such as:

			.. code-block:: pycon

				>>> URL("https://github.com/domdfcoding").path.relative_to(URL("https://github.com").path)
				Traceback (most recent call last):
				ValueError: '/domdfcoding' does not start with ''

			Since ``URL("https://github.com").path`` is ``URLPath('')``.

			Instead, use:

				>>> URL("https://github.com/domdfcoding").relative_to(URL("https://github.com"))
				URLPath('domdfcoding')
		"""

		return super().relative_to(*other)

	def as_uri(self, *args, **kwargs) -> "NoReturn":  # noqa: D102
		raise NotImplementedError


class URL(os.PathLike):
	r"""
	:mod:`pathlib`-like class for URLs.

	:param url: The URL to construct the :class:`~apeye.url.URL` object from.

	.. versionchanged:: 0.3.0  The ``url`` parameter can now be a string or a :class:`~.apeye.url.URL`.

	.. versionchanged:: 1.1.0

		Added support for sorting and rich comparisons (``<``, ``<=``, ``>`` and ``>=``).

	.. autoclasssumm:: URL
		:autosummary-sections: Methods
		:autosummary-exclude-members: __lt__,__le__,__gt__,__ge__,__init__,__hash__

	.. autosummary-widths:: 1/5

	.. autoclasssumm:: URL
		:autosummary-sections: Attributes
	"""

	#: URL scheme specifier
	scheme: str

	#: Network location part of the URL
	netloc: str

	#: The hierarchical path of the URL
	path: URLPath

	query: Dict[str, List[str]]
	"""
	The query parameters of the URL, if present.

	.. versionadded:: 0.7.0
	"""

	fragment: Optional[str]
	"""
	The URL fragment, used to identify a part of the document. :py:obj:`None` if absent from the URL.

	.. versionadded:: 0.7.0
	"""

	def __init__(self, url: Union[str, "URL"] = ''):
		if isinstance(url, URL):
			url = str(url)

		if not re.match("([A-Za-z-.]+:)?//", url):
			url = "//" + str(url)

		scheme, netloc, parts, params, query, fragment = urlparse(url)

		self.scheme: str = scheme
		self.netloc: str = netloc
		self.path = URLPath(parts)
		self.query = parse_qs(query or '')
		self.fragment = fragment or None

	@property
	def port(self) -> Optional[int]:
		"""
		The port of number of the URL as an integer, if present. Default :py:obj:`None`.

		.. versionadded:: 0.7.0
		"""

		if ':' not in self.netloc:
			return None
		else:
			return int(self.netloc.split(':')[-1])

	@classmethod
	def from_parts(
			cls: Type[URLType],
			scheme: str,
			netloc: str,
			path: PathLike,
			query: Optional[Mapping[Any, List]] = None,
			fragment: Optional[str] = None,
			) -> URLType:
		"""
		Construct a :class:`~apeye.url.URL` from a scheme, netloc and path.

		:param scheme: The scheme of the URL, e.g ``'http'``.
		:param netloc: The netloc of the URl, e.g. ``'bbc.co.uk:80'``.
		:param path: The path of the URL, e.g. ``'/news'``.
		:param query: The query parameters of the URL, if present.
		:param fragment: The URL fragment, used to identify a part of the document.
			:py:obj:`None` if absent from the URL.

		Put together, the resulting path would be ``'http://bbc.co.uk:80/news'``

		:rtype:

		.. versionchanged:: 0.7.0  Added the ``query`` and ``fragment`` arguments.
		"""

		obj = cls('')
		obj.scheme = scheme
		obj.netloc = netloc
		obj.query = dict(query or {})
		obj.fragment = fragment or None

		path = URLPath(path)

		if path.root == '/':
			obj.path = path
		else:
			obj.path = URLPath('/' + str(path))

		return obj

	def __str__(self) -> str:
		"""
		Returns the :class:`~apeye.url.URL` as a string.
		"""

		query = urlencode(self.query, doseq=True)
		url = urlunparse([self.scheme, self.netloc, str(self.path), None, query, self.fragment])

		if url.startswith("//"):
			return url[2:]
		else:
			return url

	def __repr__(self) -> str:
		"""
		Returns the string representation of the :class:`~apeye.url.URL`.
		"""

		return f"{self.__class__.__name__}({str(self)!r})"

	def __truediv__(self: URLType, key: Union[PathLike, int]) -> URLType:
		"""
		Construct a new :class:`~apeye.url.URL` object for the given child of this :class:`~apeye.url.URL`.

		:rtype:

		.. versionchanged:: 0.7.0

			* Added support for division by integers.

			* Now officially supports the new path having a URL fragment and/or query parameters.
			  Any URL fragment or query parameters from the parent URL are not inherited by its children.
		"""

		try:
			return self._make_child((key, ))
		except TypeError:
			return NotImplemented

	def _make_child(self: URLType, args: Iterable[Union[PathLike, int]]) -> URLType:
		"""
		Construct a new :class:`~apeye.url.URL` object by combining the given arguments with this instance's path part.

		.. versionadded:: 1.1.0  (private)

		Except for the final path element any queries and fragments are ignored.

		:returns: A new :class:`~.apeye.url.URL` representing either a subpath
			(if all arguments are relative paths) or a totally different path
			(if one of the arguments is absolute).
		"""

		parsed_args: List[ParseResult] = []

		for arg in args:

			raw_arg = arg

			if isinstance(arg, pathlib.PurePath):
				arg = arg.as_posix()
			elif isinstance(arg, os.PathLike):
				arg = os.fspath(arg)
			elif isinstance(arg, int):
				arg = str(arg)

			try:
				parse_result = urlparse(arg)
			except AttributeError as e:
				if str(e).endswith("'decode'"):
					msg = f"Cannot join {type(raw_arg).__name__!r} to a {type(self.path).__name__!r}"
					raise TypeError(msg) from None
				else:
					raise

			parsed_args.append(parse_result)

		try:
			new_path = self.from_parts(
					self.scheme,
					self.netloc,
					self.path.joinpath(*map(attrgetter("path"), parsed_args)),
					)
		except TypeError:
			return NotImplemented

		if parsed_args:
			new_path.query = parse_qs(parsed_args[-1].query)
			new_path.fragment = parsed_args[-1].fragment or None

		return new_path

	def joinurl(self: URLType, *args) -> URLType:
		"""
		Construct a new :class:`~apeye.url.URL` object by combining the given arguments with this instance's path part.

		.. versionadded:: 1.1.0

		Except for the final path element any queries and fragments are ignored.

		:returns: A new :class:`~.apeye.url.URL` representing either a subpath
			(if all arguments are relative paths) or a totally different path
			(if one of the arguments is absolute).
		"""

		return self._make_child(args)

	def __fspath__(self) -> str:
		"""
		Returns the file system path representation of the :class:`~.apeye.url.URL`.

		This is comprised of the ``netloc`` and ``path`` attributes.
		"""

		return f"{self.netloc}{self.path}"

	def __eq__(self, other) -> bool:
		"""
		Return ``self == other``.

		.. latex:vspace:: -10px

		.. attention::

			URL fragments and query parameters are not compared.

			.. seealso:: :meth:`.URL.strict_compare`, which *does* consider those attributes.

		.. latex:vspace:: -20px
		"""

		if isinstance(other, URL):
			return self.netloc == other.netloc and self.scheme == other.scheme and self.path == other.path
		else:
			return NotImplemented

	def __lt__(self, other):
		if isinstance(other, URL):
			return self._parts_port < other._parts_port
		else:
			return NotImplemented

	def __le__(self, other):
		if isinstance(other, URL):
			return self._parts_port <= other._parts_port
		else:
			return NotImplemented

	def __gt__(self, other):
		if isinstance(other, URL):
			return self._parts_port > other._parts_port
		else:
			return NotImplemented

	def __ge__(self, other):
		if isinstance(other, URL):
			return self._parts_port >= other._parts_port
		else:
			return NotImplemented

	def strict_compare(self, other) -> bool:
		"""
		Return ``self ≡ other``, comparing the scheme, netloc, path, fragment and query parameters.

		.. versionadded:: 0.7.0
		"""

		if isinstance(other, URL):
			return (
					self.netloc == other.netloc and self.scheme == other.scheme and self.path == other.path
					and self.query == other.query and self.fragment == other.fragment
					)
		else:
			return NotImplemented

	def __hash__(self) -> int:
		"""
		Returns the has of the :class:`~apeye.url.URL` .
		"""

		return hash((self.scheme, self.netloc, self.path))

	@property
	def name(self) -> str:
		"""
		The final path component, if any.
		"""

		return self.path.name

	@property
	def suffix(self) -> str:
		"""
		The final component's last suffix, if any.

		This includes the leading period. For example: ``'.txt'``.
		"""
		return self.path.suffix

	@property
	def suffixes(self) -> List[str]:
		"""
		A list of the final component's suffixes, if any.

		These include the leading periods. For example: ``['.tar', '.gz']``.
		"""
		return self.path.suffixes

	@property
	def stem(self):
		"""
		The final path component, minus its last suffix.
		"""

		return self.path.stem

	def with_name(self: URLType, name: str, inherit: bool = True) -> URLType:
		"""
		Return a new :class:`~apeye.url.URL` with the file name changed.

		:param name:
		:param inherit: Whether the new :class:`~apeye.url.URL` should inherit the query string
			and fragment from this :class:`~apeye.url.URL`.

		:rtype:

		.. versionchanged:: 0.7.0  Added the ``inherit`` parameter.
		"""

		if inherit:
			kwargs = {"query": self.query, "fragment": self.fragment}
		else:
			kwargs = {}

		return self.from_parts(
				self.scheme,
				self.netloc,
				self.path.with_name(name),
				**kwargs,  # type: ignore
				)

	def with_suffix(self: URLType, suffix: str, inherit: bool = True) -> URLType:
		"""
		Returns a new :class:`~apeye.url.URL` with the file suffix changed.

		If the :class:`~apeye.url.URL` has no suffix, add the given suffix.

		If the given suffix is an empty string, remove the suffix from the :class:`~apeye.url.URL`.

		:param suffix:
		:param inherit: Whether the new :class:`~apeye.url.URL` should inherit the query string
			and fragment from this :class:`~apeye.url.URL`.

		:rtype:

		.. versionchanged:: 0.7.0  Added the ``inherit`` parameter.
		"""

		if inherit:
			kwargs = {"query": self.query, "fragment": self.fragment}
		else:
			kwargs = {}

		return self.from_parts(
				self.scheme,
				self.netloc,
				self.path.with_suffix(suffix),
				**kwargs,  # type: ignore
				)

	@property
	def parts(self) -> Tuple[str, ...]:
		"""
		An object providing sequence-like access to the components in the URL.

		To retrieve only the parts of the path, use :meth:`URL.path.parts <URLPath.parts>`.
		"""

		return (
				self.scheme,
				self.domain.subdomain,
				self.domain.domain,
				self.domain.suffix,
				*('/' / self.path).parts[1:],
				)

	@property
	def _parts_port(self) -> Tuple:
		"""
		An object providing sequence-like access to the components in the URL.

		Unlike ``.parts`` this includes the port.

		To retrieve only the parts of the path, use :meth:`URL.path.parts <URLPath.parts>`.

		.. versionadded:: 1.1.0  (private)
		"""

		return (
				self.scheme,
				self.domain.subdomain,
				self.domain.domain,
				self.domain.suffix,
				self.port or 0,
				*('/' / self.path).parts[1:],
				)

	@property
	def parent(self: URLType) -> URLType:
		"""
		The logical parent of the :class:`~apeye.url.URL`.
		"""

		return self.from_parts(self.scheme, self.netloc, self.path.parent)

	@property
	def parents(self: URLType) -> Tuple[URLType, ...]:
		"""
		An immutable sequence providing access to the logical ancestors of the :class:`~apeye.url.URL`.
		"""

		return tuple(self.from_parts(self.scheme, self.netloc, path) for path in self.path.parents)

	@property
	def fqdn(self) -> str:
		"""
		Returns the Fully Qualified Domain Name of the :class:`~apeye.url.URL` .
		"""

		return self.domain.fqdn

	@property
	def domain(self) -> "Domain":
		"""
		Returns a :class:`apeye.url.Domain` object representing the domain part of the URL.
		"""

		return Domain._make(_tld.extract_tld(self.netloc))

	@property
	def base_url(self: URLType) -> URLType:
		"""
		Returns a :class:`apeye.url.URL` object representing the URL without query strings or URL fragments.

		.. versionadded:: 0.7.0
		"""

		return self.from_parts(
				self.scheme,
				self.netloc,
				self.path,
				)

	def relative_to(self, other: Union[str, "URL", URLPath]) -> URLPath:
		"""
		Returns a version of this URL's path relative to ``other``.

		.. versionadded:: 1.1.0

		:param other: Either a :class:`~.apeye.url.URL`, or a string or :class:`~.apeye.url.URLPath` representing an *absolute* path.
			If a :class:`~.apeye.url.URL`, the :attr:`~.apeye.url.URL.netloc` must match this URL's.

		:raises ValueError: if the operation is not possible
			(i.e. because this URL's path is not a subpath of the other path)
		"""

		if isinstance(other, URLPath):
			if not other.is_absolute():
				raise ValueError("'URL.relative_to' cannot be used with relative URLPath objects")
			else:
				other = URL('/') / other
		elif not isinstance(other, URL):
			# Parse other as a URL
			other = URL(other)

		# Compare netloc, if both have one
		if self.netloc and other.netloc and self.netloc.lower() != other.netloc.lower():
			raise ValueError(f"{self!r} does not start with {other!r}")

		# Make the paths absolute
		# If coming from a URL they must always be absolute
		our_path = '/' / self.path
		other_path = '/' / other.path

		relative_path = our_path.relative_to(other_path)

		return relative_path


class Domain(NamedTuple):
	"""
	:class:`typing.NamedTuple` of a URL's subdomain, domain, and suffix.
	"""

	subdomain: str
	domain: str
	suffix: str

	@property
	def registered_domain(self):
		"""
		Joins the domain and suffix fields with a dot, if they're both set.

		.. code-block:: python

			>>> URL('https://forums.bbc.co.uk').domain.registered_domain
			'bbc.co.uk'
			>>> URL('https://localhost:8080').domain.registered_domain
			''
		"""
		if self.domain and self.suffix:
			return self.domain + '.' + self.suffix
		return ''

	@property
	def fqdn(self):
		"""
		Returns a Fully Qualified Domain Name, if there is a proper domain/suffix.

		.. code-block:: python

			>>> URL('https://forums.bbc.co.uk/path/to/file').domain.fqdn
			'forums.bbc.co.uk'
			>>> URL('https://localhost:8080').domain.fqdn
			''
		"""
		if self.domain and self.suffix:
			# self is the namedtuple (subdomain domain suffix)
			return '.'.join(i for i in self if i)
		return ''

	@property
	def ipv4(self) -> Optional[ipaddress.IPv4Address]:
		"""
		Returns the ipv4 if that is what the presented domain/url is.

		.. code-block:: python

			>>> URL('https://127.0.0.1/path/to/file').domain.ipv4
			IPv4Address('127.0.0.1')
			>>> URL('https://127.0.0.1.1/path/to/file').domain.ipv4
			>>> URL('https://256.1.1.1').domain.ipv4
		"""

		if not (self.suffix or self.subdomain) and _tld.IP_RE.match(self.domain):
			return cast(ipaddress.IPv4Address, ipaddress.ip_address(self.domain))
		return None

	def __repr__(self) -> str:
		"""
		Return a string representation of the :class:`~.Domain`.
		"""

		# This is necessary to get the custom docstring

		repr_fmt = f"({', '.join(f'{name}=%r' for name in self._fields)})"
		return f"{self.__class__.__name__}{repr_fmt % self}"



#!/usr/bin/env python
#
#  _tld.py
"""
Functions for parsing top-level domains (TLDs).

A stripped back version of https://github.com/john-kurkowski/tldextract
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Adapted from https://github.com/john-kurkowski/tldextract
#  Licensed under the BSD 3-Clause License
#
#  Copyright (c) 2020, John Kurkowski
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
import functools
import re
import socket
from typing import List, Tuple
from urllib.parse import scheme_chars

# 3rd party
import idna
from domdf_python_tools.compat import importlib_resources

__all__ = [
		"determine_suffix_index",
		"extract_tld",
		"extract_tlds_from_suffix_list",
		"load_suffix_list",
		"looks_like_ip"
		]

PUBLIC_SUFFIX_RE = re.compile(r"^(?P<suffix>[.*!]*\w[\S]*)", re.UNICODE | re.MULTILINE)
PUBLIC_PRIVATE_SUFFIX_SEPARATOR: str = "// ===BEGIN PRIVATE DOMAINS==="
IP_RE = re.compile(r'^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$')  # pylint: disable=line-too-long
SCHEME_RE = re.compile(fr'^([{scheme_chars}]+:)?//')


def extract_tlds_from_suffix_list(suffix_list_text: str) -> Tuple[List[str], List[str]]:
	"""
	Parse the raw suffix list text for its different designations of suffixes.

	:param suffix_list_text:

	:returns: A two element tuple of ``(public TLDs, private TLDs)``.
	"""

	public_text, _, private_text = suffix_list_text.partition(PUBLIC_PRIVATE_SUFFIX_SEPARATOR)
	public_tlds = [m.group("suffix") for m in PUBLIC_SUFFIX_RE.finditer(public_text)]
	private_tlds = [m.group("suffix") for m in PUBLIC_SUFFIX_RE.finditer(private_text)]
	return public_tlds, private_tlds


@functools.lru_cache(1)
def load_suffix_list() -> Tuple[List[str], List[str]]:
	"""
	Load the public suffix list from disk.

	:returns: A two element tuple of ``(public TLDs, private TLDs)``.
	"""

	return extract_tlds_from_suffix_list(importlib_resources.read_text("apeye_core", "public_suffix_list.dat"))


def _decode_punycode(label: str) -> str:
	lowered = label.lower()
	looks_like_puny = lowered.startswith("xn--")

	if looks_like_puny:
		try:
			return idna.decode(label.encode("ascii")).lower()
		except (UnicodeError, IndexError):
			pass

	return lowered


def determine_suffix_index(tlds: List[str], elements: List[str]) -> int:
	"""
	Returns the index of the first suffix label, or ``len(elements)`` if no suffix is found.

	:param tlds: The list of public TLDs.
	:param elements: The elements of the domain, e.g. ``['forums', 'bbc', 'co', 'uk']``.
	"""

	length = len(elements)
	for i in range(length):
		maybe_tld = '.'.join(elements[i:])
		exception_tld = '!' + maybe_tld

		if exception_tld in tlds:
			return i + 1

		if maybe_tld in tlds:
			return i

		wildcard_tld = "*." + '.'.join(elements[i + 1:])
		if wildcard_tld in tlds:
			return i

	return length


def looks_like_ip(maybe_ip: str) -> bool:
	"""
	Returns whether the given string looks like an IP address.

	:param maybe_ip:
	"""

	if not maybe_ip[0].isdigit():
		return False

	try:
		socket.inet_aton(maybe_ip)
		return True

	except (AttributeError, UnicodeError):
		if IP_RE.match(maybe_ip):
			return True
		return False

	except OSError:
		return False


def extract_tld(url: str) -> Tuple[str, str, str]:
	"""
	Takes a string URL and splits it into its subdomain, domain, and suffix
	(effective TLD, gTLD, ccTLD, etc.) component.

	.. versionadded:: 1.0.0 (undocumented)

	.. code-block:: python

		>>> Domain._make(extract_tld('https://forums.news.cnn.com/'))
		Domain(subdomain='forums.news', domain='cnn', suffix='com')
		>>> Domain._make(extract_tld('https://forums.bbc.co.uk/')
		Domain(subdomain='forums', domain='bbc', suffix='co.uk')
	"""  # noqa: D400

	netloc_ = SCHEME_RE.sub('', url).partition('/')[0].partition('?')[0].partition('#')[0]
	netloc = netloc_.split('@')[-1].partition(':')[0].strip().rstrip('.')

	labels = netloc.split('.')

	translations = [_decode_punycode(label) for label in labels]
	tlds, _ = load_suffix_list()
	suffix_index = determine_suffix_index(tlds, translations)
	suffix = '.'.join(labels[suffix_index:])

	if not suffix and netloc and looks_like_ip(netloc):
		return '', netloc, ''

	subdomain = '.'.join(labels[:suffix_index - 1]) if suffix_index else ''
	domain = labels[suffix_index - 1] if suffix_index else ''
	return subdomain, domain, suffix



#!/usr/bin/env python
#
#  url.py
"""
Email address validation functions.

.. versionadded:: 1.0.0

This module is a subset of https://pypi.org/project/email-validator
"""
#
# Based on https://github.com/JoshData/python-email-validator
# Licensed under the CC0 License
#

# stdlib
import re
import sys
import unicodedata
from typing import Any, Dict, Optional, Union

# 3rd party
import idna  # implements IDNA 2008; Python's codec is only IDNA 2003
from domdf_python_tools.doctools import prettify_docstrings

__all__ = [
		"EmailSyntaxError",
		"ValidatedEmail",
		"main",
		"validate_email",
		"validate_email_domain_part",
		"validate_email_local_part"
		]

# Based on RFC 2822 section 3.2.4 / RFC 5322 section 3.2.3, these
# characters are permitted in email addresses (not taking into
# account internationalization):
ATEXT: str = r'a-zA-Z0-9_!#\$%&\'\*\+\-/=\?\^`\{\|\}~'

# A "dot atom text", per RFC 2822 3.2.4:
DOT_ATOM_TEXT: str = f'[{ATEXT}]+(?:\\.[{ATEXT}]+)*'

# RFC 6531 section 3.3 extends the allowed characters in internationalized
# addresses to also include three specific ranges of UTF8 defined in
# RFC3629 section 4, which appear to be the Unicode code points from
# U+0080 to U+10FFFF.
ATEXT_UTF8: str = ATEXT + "-􏿿"  # Don't convert to f-string
DOT_ATOM_TEXT_UTF8: str = f'[{ATEXT_UTF8}]+(?:\\.[{ATEXT_UTF8}]+)*'

# The domain part of the email address, after IDNA (ASCII) encoding,
# must also satisfy the requirements of RFC 952/RFC 1123 which restrict
# the allowed characters of hostnames further. The hyphen cannot be at
# the beginning or end of a *dot-atom component* of a hostname either.
ATEXT_HOSTNAME: str = r'(?:(?:[a-zA-Z0-9][a-zA-Z0-9\-]*)?[a-zA-Z0-9])'

# Length constants
# RFC 3696 + errata 1003 + errata 1690 (https://www.rfc-editor.org/errata_search.php?rfc=3696&eid=1690)
# explains the maximum length of an email address is 254 octets.
EMAIL_MAX_LENGTH: int = 254
LOCAL_PART_MAX_LENGTH: int = 64
DOMAIN_MAX_LENGTH: int = 255


class EmailSyntaxError(ValueError):
	"""
	Exception raised when an email address fails validation because of its form.
	"""


@prettify_docstrings
class ValidatedEmail:
	"""
	Represents the return type of the :func:`~apeye.email_validator.validate_email` function.

	This class holds the normalized form of the email address alongside other information.

	:param original_email: The original, unnormalized email address.
	:param email: The normalized email address, which should always be used in preference to the original address.
	:param local_part: The local part of the email address after Unicode normalization.
	:param domain: The domain part of the email address after Unicode normalization or conversion to Unicode from IDNA ascii.
	:param ascii_email: If not :py:obj:`None`, a form of the email address that uses 7-bit ASCII characters only.
	:param ascii_local_part: If not :py:obj:`None`, the local part of the email address using 7-bit ASCII characters only.
	:param ascii_domain: If not :py:obj:`None`, a form of the domain name that uses 7-bit ASCII characters only.
	:param smtputf8: Indicates whether SMTPUTF8 will be required to transmit messages to this address.

	.. autosummary-widths:: 1/4
	"""

	#: The email address that was passed to validate_email. (If passed as bytes, this will be a string.)
	original_email: str

	email: str
	"""
	The normalized email address, which should always be used in preference to the original address.

	The normalized address converts an IDNA ASCII domain name to Unicode, if possible,
	and performs Unicode normalization on the local part and on the domain (if originally Unicode).
	It is the concatenation of the local_part and domain attributes, separated by an @-sign.
	"""

	#: The local part of the email address after Unicode normalization.
	local_part: str

	#: The domain part of the email address after Unicode normalization or conversion to Unicode from IDNA ascii.
	domain: str

	#: If not :py:obj:`None`, a form of the email address that uses 7-bit ASCII characters only.
	ascii_email: Optional[str]

	#: If not :py:obj:`None`, the local part of the email address using 7-bit ASCII characters only.
	ascii_local_part: Optional[str]

	#: If not :py:obj:`None`, a form of the domain name that uses 7-bit ASCII characters only.
	ascii_domain: Optional[str]

	smtputf8: Optional[bool]
	"""
	If :py:obj:`True`, the SMTPUTF8 feature of your mail relay will be required to transmit messages
	to this address.

	This flag is :py:obj:`True` when :attr:`~.ascii_local_part` is missing.
	Otherwise it is :py:obj:`False`.
	"""

	def __init__(
			self,
			original_email: str,
			email: str,
			local_part: str,
			domain: str,
			*,
			ascii_email: Optional[str] = None,
			ascii_local_part: Optional[str] = None,
			ascii_domain: Optional[str] = None,
			smtputf8: Optional[bool] = None,
			):
		self.original_email = original_email
		self.email = email
		self.local_part = local_part
		self.domain = domain
		self.ascii_email = ascii_email
		self.ascii_local_part = ascii_local_part
		self.ascii_domain = ascii_domain
		self.smtputf8 = smtputf8

	def __str__(self) -> str:
		"""
		Return a string representation of the :class:`~apeye.email_validator.ValidatedEmail` object.
		"""

		return self.email

	def __repr__(self) -> str:
		"""
		Return a string representation of the :class:`~apeye.email_validator.ValidatedEmail` object.
		"""

		return f"<ValidatedEmail {self.email}>"

	def __eq__(self, other) -> bool:
		"""
		Return ``self == other``.
		"""

		return (
				self.email == other.email and self.local_part == other.local_part and self.domain == other.domain
				and self.ascii_email == other.ascii_email and self.ascii_local_part == other.ascii_local_part
				and self.ascii_domain == other.ascii_domain and self.smtputf8 == other.smtputf8
				)

	def as_dict(self) -> Dict[str, Any]:
		"""
		Convenience method for accessing the :class:`~apeye.email_validator.ValidatedEmail` as a dict.
		"""

		return {
				"original_email": self.original_email,
				"email": self.email,
				"local_part": self.local_part,
				"domain": self.domain,
				"ascii_email": self.ascii_email,
				"ascii_local_part": self.ascii_local_part,
				"ascii_domain": self.ascii_domain,
				"smtputf8": self.smtputf8,
				}


def __get_length_reason(addr: str, utf8: bool = False, limit: int = EMAIL_MAX_LENGTH):
	diff = len(addr) - limit
	prefix = "at least " if utf8 else ''
	suffix = 's' if diff > 1 else ''
	return f"({prefix}{diff} character{suffix} too many)"


def validate_email(
		email: Union[str, bytes],
		allow_smtputf8: bool = True,
		allow_empty_local: bool = False,
		) -> ValidatedEmail:
	"""
	Validates an email address.

	:param email: Either a string, or ASCII-encoded bytes.
	:param allow_smtputf8:
	:param allow_empty_local: Whether to allow the local part (the bit before the @-sign) to be empty.

	:raises EmailSyntaxError: if the address is not valid
	"""

	# Allow email to be a str or bytes instance. If bytes,
	# it must be ASCII because that's how the bytes work
	# on the wire with SMTP.
	if not isinstance(email, str):
		try:
			email = email.decode("ascii")
		except ValueError:
			raise EmailSyntaxError("The email address is not valid ASCII.")

	# At-sign.
	parts = email.split('@')
	if len(parts) != 2:
		raise EmailSyntaxError("The email address is not valid. It must have exactly one @-sign.")

	# Validate the email address's local part syntax and get a normalized form.
	local_part_info = validate_email_local_part(
			parts[0],
			allow_smtputf8=allow_smtputf8,
			allow_empty_local=allow_empty_local,
			)

	# Validate the email address's domain part syntax and get a normalized form.
	domain_part_info = validate_email_domain_part(parts[1])

	ret = ValidatedEmail(
			original_email=email,
			email=f"{local_part_info['local_part']}@{domain_part_info['domain']}",
			local_part=local_part_info["local_part"],
			domain=domain_part_info["domain"],
			ascii_local_part=local_part_info["ascii_local_part"],
			smtputf8=local_part_info["smtputf8"],
			ascii_domain=domain_part_info["ascii_domain"],
			)

	# If the email address has an ASCII form, add it.
	if not ret.smtputf8:
		ret.ascii_email = ret.ascii_local_part + '@' + ret.ascii_domain  # type: ignore[operator]

	# If the email address has an ASCII representation, then we assume it may be
	# transmitted in ASCII (we can't assume SMTPUTF8 will be used on all hops to
	# the destination) and the length limit applies to ASCII characters (which is
	# the same as octets). The number of characters in the internationalized form
	# may be many fewer (because IDNA ASCII is verbose) and could be less than 254
	# Unicode characters, and of course the number of octets over the limit may
	# not be the number of characters over the limit, so if the email address is
	# internationalized, we can't give any simple information about why the address
	# is too long.
	#
	# In addition, check that the UTF-8 encoding (i.e. not IDNA ASCII and not
	# Unicode characters) is at most 254 octets. If the addres is transmitted using
	# SMTPUTF8, then the length limit probably applies to the UTF-8 encoded octets.
	# If the email address has an ASCII form that differs from its internationalized
	# form, I don't think the internationalized form can be longer, and so the ASCII
	# form length check would be sufficient. If there is no ASCII form, then we have
	# to check the UTF-8 encoding. The UTF-8 encoding could be up to about four times
	# longer than the number of characters.
	#
	# See the length checks on the local part and the domain.

	if ret.ascii_email and len(ret.ascii_email) > EMAIL_MAX_LENGTH:
		if ret.ascii_email == ret.email:
			reason = __get_length_reason(ret.ascii_email)
		elif len(ret.email) > EMAIL_MAX_LENGTH:
			# If there are more than 254 characters, then the ASCII
			# form is definitely going to be too long.
			reason = __get_length_reason(ret.email, utf8=True)
		else:
			reason = "(when converted to IDNA ASCII)"
		raise EmailSyntaxError(f"The email address is too long {reason}.")

	if len(ret.email.encode("utf8")) > EMAIL_MAX_LENGTH:
		if len(ret.email) > EMAIL_MAX_LENGTH:
			# If there are more than 254 characters, then the UTF-8
			# encoding is definitely going to be too long.
			reason = __get_length_reason(ret.email, utf8=True)
		else:
			reason = "(when encoded in bytes)"
		raise EmailSyntaxError(f"The email address is too long {reason}.")

	return ret


def validate_email_local_part(
		local: str,
		allow_smtputf8: bool = True,
		allow_empty_local: bool = False,
		) -> Dict[str, Any]:
	"""
	Validates the local part of an email address  (the part before the @-sign).

	:param local:
	:param allow_smtputf8:
	:param allow_empty_local: Whether to allow the local part to be empty/
	"""

	if len(local) == 0:
		if not allow_empty_local:
			raise EmailSyntaxError("There must be something before the @-sign.")
		else:
			# The caller allows an empty local part. Useful for validating certain
			# Postfix aliases.
			return {
					"local_part": local,
					"ascii_local_part": local,
					"smtputf8": False,
					}

	# RFC 5321 4.5.3.1.1
	# We're checking the number of characters here. If the local part
	# is ASCII-only, then that's the same as bytes (octets). If it's
	# internationalized, then the UTF-8 encoding may be longer, but
	# that may not be relevant. We will check the total address length
	# instead.
	if len(local) > LOCAL_PART_MAX_LENGTH:
		reason = __get_length_reason(local, limit=LOCAL_PART_MAX_LENGTH)
		raise EmailSyntaxError(f"The email address is too long before the @-sign {reason}.")

	# Check the local part against the regular expression for the older ASCII requirements.
	m = re.match(DOT_ATOM_TEXT + "\\Z", local)
	if m:
		# Return the local part unchanged and flag that SMTPUTF8 is not needed.
		return {
				"local_part": local,
				"ascii_local_part": local,
				"smtputf8": False,
				}

	else:
		# The local part failed the ASCII check. Now try the extended internationalized requirements.
		m = re.match(DOT_ATOM_TEXT_UTF8 + "\\Z", local)
		if not m:
			# It's not a valid internationalized address either. Report which characters were not valid.
			bad_chars = ", ".join(
					sorted({
							c
							for c in local
							if not re.match('[' + (ATEXT if not allow_smtputf8 else ATEXT_UTF8) + ']', c)
							})
					)
			raise EmailSyntaxError(
					f"The email address contains invalid characters before the @-sign: {bad_chars}."
					)

		# It would be valid if internationalized characters were allowed by the caller.
		if not allow_smtputf8:
			raise EmailSyntaxError("Internationalized characters before the @-sign are not supported.")

		# It's valid.

		# RFC 6532 section 3.1 also says that Unicode NFC normalization should be applied,
		# so we'll return the normalized local part in the return value.
		local = unicodedata.normalize("NFC", local)

		# Flag that SMTPUTF8 will be required for deliverability.
		return {
				"local_part": local,
				"ascii_local_part": None,  # no ASCII form is possible
				"smtputf8": True,
				}


def validate_email_domain_part(domain: str) -> Dict[str, str]:
	"""
	Validate the domain part of an email address (the part after the @-sign).

	:param domain:
	"""

	# Empty?
	if len(domain) == 0:
		raise EmailSyntaxError("There must be something after the @-sign.")

	# Perform UTS-46 normalization, which includes casefolding, NFC normalization,
	# and converting all label separators (the period/full stop, fullwidth full stop,
	# ideographic full stop, and halfwidth ideographic full stop) to basic periods.
	# It will also raise an exception if there is an invalid character in the input,
	# such as "⒈" which is invalid because it would expand to include a period.
	try:
		domain = idna.uts46_remap(domain, std3_rules=False, transitional=False)
	except idna.IDNAError as e:
		raise EmailSyntaxError(f"The domain name {domain} contains invalid characters ({str(e)}).")

	# Now we can perform basic checks on the use of periods (since equivalent
	# symbols have been mapped to periods). These checks are needed because the
	# IDNA library doesn't handle well domains that have empty labels (i.e. initial
	# dot, trailing dot, or two dots in a row).
	if domain.endswith('.'):
		raise EmailSyntaxError("An email address cannot end with a period.")
	if domain.startswith('.'):
		raise EmailSyntaxError("An email address cannot have a period immediately after the @-sign.")
	if ".." in domain:
		raise EmailSyntaxError("An email address cannot have two periods in a row.")

	# Regardless of whether international characters are actually used,
	# first convert to IDNA ASCII. For ASCII-only domains, the transformation
	# does nothing. If internationalized characters are present, the MTA
	# must either support SMTPUTF8 or the mail client must convert the
	# domain name to IDNA before submission.
	#
	# Unfortunately this step incorrectly 'fixes' domain names with leading
	# periods by removing them, so we have to check for this above. It also gives
	# a funky error message ("No input") when there are two periods in a
	# row, also checked separately above.
	try:
		ascii_domain = idna.encode(domain, uts46=False).decode("ascii")
	except idna.IDNAError as e:
		if "Domain too long" in str(e):
			# We can't really be more specific because UTS-46 normalization means
			# the length check is applied to a string that is different from the
			# one the user supplied. Also I'm not sure if the length check applies
			# to the internationalized form, the IDNA ASCII form, or even both!
			raise EmailSyntaxError("The email address is too long after the @-sign.")
		raise EmailSyntaxError(f"The domain name {domain} contains invalid characters ({str(e)}).")

	# We may have been given an IDNA ASCII domain to begin with. Check
	# that the domain actually conforms to IDNA. It could look like IDNA
	# but not be actual IDNA. For ASCII-only domains, the conversion out
	# of IDNA just gives the same thing back.
	#
	# This gives us the canonical internationalized form of the domain,
	# which we should use in all error messages.
	try:
		domain_i18n = idna.decode(ascii_domain.encode("ascii"))
	except idna.IDNAError as e:
		raise EmailSyntaxError(f"The domain name {ascii_domain} is not valid IDNA ({str(e)}).")

	# RFC 5321 4.5.3.1.2
	# We're checking the number of bytes (octets) here, which can be much
	# higher than the number of characters in internationalized domains,
	# on the assumption that the domain may be transmitted without SMTPUTF8
	# as IDNA ASCII. This is also checked by idna.encode, so this exception
	# is never reached.
	if len(ascii_domain) > DOMAIN_MAX_LENGTH:
		raise EmailSyntaxError("The email address is too long after the @-sign.")

	# A "dot atom text", per RFC 2822 3.2.4, but using the restricted
	# characters allowed in a hostname (see ATEXT_HOSTNAME above).
	DOT_ATOM_TEXT = ATEXT_HOSTNAME + r'(?:\.' + ATEXT_HOSTNAME + r')*'

	# Check the regular expression. This is probably entirely redundant
	# with idna.decode, which also checks this format.
	m = re.match(DOT_ATOM_TEXT + "\\Z", ascii_domain)
	if not m:
		raise EmailSyntaxError("The email address contains invalid characters after the @-sign.")

	# All publicly deliverable addresses have domain named with at least
	# one period. We also know that all TLDs end with a letter.
	if '.' not in ascii_domain:
		raise EmailSyntaxError("The domain name %s is not valid. It should have a period." % domain_i18n)
	if not re.search(r"[A-Za-z]\Z", ascii_domain):
		raise EmailSyntaxError(
				"The domain name %s is not valid. It is not within a valid top-level domain." % domain_i18n
				)

	# Return the IDNA ASCII-encoded form of the domain, which is how it
	# would be transmitted on the wire (except when used with SMTPUTF8
	# possibly), as well as the canonical Unicode form of the domain,
	# which is better for display purposes. This should also take care
	# of RFC 6532 section 3.1's suggestion to apply Unicode NFC
	# normalization to addresses.
	return {
			"ascii_domain": ascii_domain,
			"domain": domain_i18n,
			}


def main():  # noqa: D103
	# stdlib
	import json

	if len(sys.argv) == 1:
		for line in sys.stdin:
			email = line.strip()
			try:
				validate_email(email)
			except EmailSyntaxError as e:
				print(f"{email} {e}")
	else:
		# Validate the email address passed on the command line.
		email = sys.argv[1]
		try:
			result = validate_email(email)
			print(json.dumps(result.as_dict(), indent=2, sort_keys=True, ensure_ascii=False))
		except EmailSyntaxError as e:
			print(e)


if __name__ == "__main__":
	sys.exit(main())
